%%%%%%%%%%%%%%%%%%%%%%%%%%%% PATH PLANNING FUNCTIONS %%%%%%%%%%%%%%%%%%%%%%%%%%%

% This function is called from the main loop in arw_simulator
% it calls the RRT planner to build a collision-free path
function [T, nodes] = constructPath(X_est, X_des, M, T, nodes)

    % Only get a new path if one has not been made
    if size(T,1) < 1
        [T, nodes] = buildRRT(X_est, X_des, M);
    end

end




%This is the core RRT motion planning algorithm to be written by ARW
%participants.
function [T, nodes] = buildRRT(X_start, X_goal, M)

    % Create NodeList
    nodes = [X_start(1) X_start(2) 0];
	pathFound = false;
	maxIterations = 10000;
	iterations = 0;
    mapRanges = getMapRanges(M);

	

    
    % Expand the tree until a path is found OR too many iterations have
    % passed.
	while iterations < maxIterations && pathFound == false

		% Choose node for expansion, set nodeToExpand = ...
        
        
        
        
        
        
		% Expand Node, set newNode = .....
        
        
		% Check for collision, add newNode and check if connected to goal

         
        
        
		% Increment number of iterations
		iterations = iterations + 1;
    end
    
    % Create the trajectory to follow
	T = BuildOptimalPath(nodes, mapRanges, M);
    
end



% This function is called from buildRRT to construct a path starting from
% the goal node and backtracking through parents until the start node is
% reached. Each node is entered in the path twice, one for the incoming
% angle and out for the outgoing angle.
function [T] = BuildOptimalPath(nodes, mapRanges, M)

    nextNode = nodes(size(nodes,1),1:3);
    T = [];
    while nextNode(3) > 0
        
        % Set parent, but reset if it can be skipped
        parentNode = nodes(nextNode(3),1:3);
        while parentNode(3) > 0
            grandParentNode = nodes(parentNode(3),1:3);
            if collisionFound(grandParentNode, nextNode, mapRanges, M) == false
                parentNode = grandParentNode;                 
            else
                break;
            end
        end
        
        theta = atan2(nextNode(2)-parentNode(2),nextNode(1)-parentNode(1));
        T = [nextNode(1) nextNode(2) theta 0;T];
        T = [parentNode(1) parentNode(2) theta 0;T];
        nextNode = parentNode; 
    end
    T = [nextNode(1) nextNode(2) 0 1;T];
    
   
end

% This function is queried by the rrt algorithm to determine if a potential
% new edge collides with a wall or boundary edge.
function [collision] = collisionFound(n1, n2, mapRanges, M)

	% Check that within boundaries
    bTol = 2;
	if n2(1) < mapRanges(1)-bTol || n2(1) > mapRanges(2)+bTol || n2(2) < mapRanges(3)-bTol || n2(2) > mapRanges(4)+bTol
		collision = true;
        return;
    end

	% Check for collision with walls
	theta = atan2(n2(2)-n1(2), n2(1)-n1(1));
	edgeSize = sqrt( (n2(2)-n1(2))^2 + (n2(1)-n1(1))^2);
	sinTheta = sin(theta);
	cosTheta = cos(theta);
    tol = 0.1;
    
	% Loop through segments
	for i=1:size(M,1)

		distTravelledOnEdge = 0;
		e = n1;
		while distTravelledOnEdge-tol < edgeSize;
			distToSegment = getWallDistance(e(1), e(2), M(i,1:7));
			if (distToSegment-tol < 0.05)
				collision = true;
                return;
            end
			e = [e(1)+cosTheta*distToSegment e(2)+sinTheta*distToSegment];
			distTravelledOnEdge = distTravelledOnEdge+distToSegment;
        end

        collision = false;
    end

end


% This function is called by the collision checker to see how close a
% position is to a wall.
function [dist] = getWallDistance(x, y, wall)

	% Set wall vars
	X1 = wall(1);
	Y1 = wall(2);
	X2 = wall(3);
	Y2 = wall(4);
	dist = 9999;
    tol = 0.2;

	% Put code here to calculated dist.
	% Calculate slope and intercept
	angleSegmentPerpendicular = pi/2 + atan((Y2-Y1)/(0.000001+X2-X1));
	m = tan(angleSegmentPerpendicular);
	b = y - m*x;

	% Get line intersection
    % Check if wall is horizontal
    if m > 9999
        x_intersect = x;
        y_intersect = Y1;
    else
        x_intersect = (b-wall(7))/(wall(6)-m);
        y_intersect = m*x_intersect + b;
    end
    
	% Check to see if intersection LIES within segment
	dist_intersect_corner1 = sqrt( (x_intersect-X1)^2 + (y_intersect-Y1)^2);
	dist_intersect_corner2 = sqrt( (x_intersect-X2)^2 + (y_intersect-Y2)^2);
	if dist_intersect_corner1 <= (wall(5)+tol) && dist_intersect_corner2 <= (wall(5)+tol) 
		dist = sqrt( (x-x_intersect)^2 + (y-y_intersect)^2);
    end
	
	% Check for distance to corners (for case where no intersection with segment
	dist_point_corner1 = sqrt( (x-X1)^2 + (y-Y1)^2);
	dist_point_corner2 = sqrt( (x-X2)^2 + (y-Y2)^2);
	dist = min(dist, dist_point_corner1); 
	dist = min(dist, dist_point_corner2); 
    
end

% This function finds the min and max values of x,y coordinates of wall
% endpoints.
function [mapRanges] = getMapRanges(M)
      
    minX = min(min(M(:,1)),min(M(:,3)));
    maxX = max(max(M(:,1)),min(M(:,3)));
    minY = min(min(M(:,2)),max(M(:,4)));
    maxY = max(max(M(:,2)),max(M(:,4)));
    mapRanges = [minX maxX minY maxY];
end

